@@ -7,7 +7,7 @@ from rewrite.java import *
 from rewrite.java.support_types import *
 from rewrite.java.visitor import JavaVisitor
 
-from rewrite.remote import Receiver, ReceiverContext, T, ReceiverFactory
+from rewrite_remote import Receiver, ReceiverContext, T, ReceiverFactory
 from . import extensions
 
 class JavaReceiver(Receiver):
@@ -124,9 +124,10 @@ class JavaReceiver(Receiver):
             case = case.with_prefix(ctx.receive_node(case.prefix, JavaReceiver.receive_space))
             case = case.with_markers(ctx.receive_node(case.markers, ctx.receive_markers))
             case = case.with_type(ctx.receive_value(case.type, Case.Type))
-            case = case.padding.with_expressions(ctx.receive_node(case.padding.expressions, JavaReceiver.receive_container))
+            case = case.padding.with_case_labels(ctx.receive_node(case.padding.case_labels, JavaReceiver.receive_container))
             case = case.padding.with_statements(ctx.receive_node(case.padding.statements, JavaReceiver.receive_container))
             case = case.padding.with_body(ctx.receive_node(case.padding.body, JavaReceiver.receive_right_padded_tree))
+            case = case.with_guard(ctx.receive_node(case.guard, ctx.receive_tree))
             return case
 
         def visit_class_declaration(self, class_declaration: ClassDeclaration, ctx: ReceiverContext) -> J:
@@ -302,6 +303,15 @@ class JavaReceiver(Receiver):
             instance_of = instance_of.with_type(ctx.receive_value(instance_of.type, JavaType))
             return instance_of
 
+        def visit_deconstruction_pattern(self, deconstruction_pattern: DeconstructionPattern, ctx: ReceiverContext) -> J:
+            deconstruction_pattern = deconstruction_pattern.with_id(ctx.receive_value(deconstruction_pattern.id, UUID))
+            deconstruction_pattern = deconstruction_pattern.with_prefix(ctx.receive_node(deconstruction_pattern.prefix, JavaReceiver.receive_space))
+            deconstruction_pattern = deconstruction_pattern.with_markers(ctx.receive_node(deconstruction_pattern.markers, ctx.receive_markers))
+            deconstruction_pattern = deconstruction_pattern.with_deconstructor(ctx.receive_node(deconstruction_pattern.deconstructor, ctx.receive_tree))
+            deconstruction_pattern = deconstruction_pattern.padding.with_nested(ctx.receive_node(deconstruction_pattern.padding.nested, JavaReceiver.receive_container))
+            deconstruction_pattern = deconstruction_pattern.with_type(ctx.receive_value(deconstruction_pattern.type, JavaType))
+            return deconstruction_pattern
+
         def visit_intersection_type(self, intersection_type: IntersectionType, ctx: ReceiverContext) -> J:
             intersection_type = intersection_type.with_id(ctx.receive_value(intersection_type.id, UUID))
             intersection_type = intersection_type.with_prefix(ctx.receive_node(intersection_type.prefix, JavaReceiver.receive_space))
@@ -496,6 +506,7 @@ class JavaReceiver(Receiver):
             switch_expression = switch_expression.with_markers(ctx.receive_node(switch_expression.markers, ctx.receive_markers))
             switch_expression = switch_expression.with_selector(ctx.receive_node(switch_expression.selector, ctx.receive_tree))
             switch_expression = switch_expression.with_cases(ctx.receive_node(switch_expression.cases, ctx.receive_tree))
+            switch_expression = switch_expression.with_type(ctx.receive_value(switch_expression.type, JavaType))
             return switch_expression
 
         def visit_synchronized(self, synchronized: Synchronized, ctx: ReceiverContext) -> J:
@@ -644,6 +655,13 @@ class JavaReceiver(Receiver):
             source = source.with_text(ctx.receive_value(source.text, str))
             return source
 
+        def visit_erroneous(self, erroneous: Erroneous, ctx: ReceiverContext) -> J:
+            erroneous = erroneous.with_id(ctx.receive_value(erroneous.id, UUID))
+            erroneous = erroneous.with_prefix(ctx.receive_node(erroneous.prefix, JavaReceiver.receive_space))
+            erroneous = erroneous.with_markers(ctx.receive_node(erroneous.markers, ctx.receive_markers))
+            erroneous = erroneous.with_text(ctx.receive_value(erroneous.text, str))
+            return erroneous
+
     # noinspection PyTypeChecker
     class Factory(ReceiverFactory):
         def create(self, type: str, ctx: ReceiverContext) -> Tree:
@@ -753,7 +771,8 @@ class JavaReceiver(Receiver):
                     ctx.receive_value(None, Case.Type),
                     ctx.receive_node(None, JavaReceiver.receive_container),
                     ctx.receive_node(None, JavaReceiver.receive_container),
-                    ctx.receive_node(None, JavaReceiver.receive_right_padded_tree)
+                    ctx.receive_node(None, JavaReceiver.receive_right_padded_tree),
+                    ctx.receive_node(None, ctx.receive_tree)
                 )
 
             if type in ["rewrite.java.tree.ClassDeclaration", "org.openrewrite.java.tree.J$ClassDeclaration"]:
@@ -948,6 +967,16 @@ class JavaReceiver(Receiver):
                     ctx.receive_value(None, JavaType)
                 )
 
+            if type in ["rewrite.java.tree.DeconstructionPattern", "org.openrewrite.java.tree.J$DeconstructionPattern"]:
+                return DeconstructionPattern(
+                    ctx.receive_value(None, UUID),
+                    ctx.receive_node(None, JavaReceiver.receive_space),
+                    ctx.receive_node(None, ctx.receive_markers),
+                    ctx.receive_node(None, ctx.receive_tree),
+                    ctx.receive_node(None, JavaReceiver.receive_container),
+                    ctx.receive_value(None, JavaType)
+                )
+
             if type in ["rewrite.java.tree.IntersectionType", "org.openrewrite.java.tree.J$IntersectionType"]:
                 return IntersectionType(
                     ctx.receive_value(None, UUID),
@@ -1163,7 +1192,8 @@ class JavaReceiver(Receiver):
                     ctx.receive_node(None, JavaReceiver.receive_space),
                     ctx.receive_node(None, ctx.receive_markers),
                     ctx.receive_node(None, ctx.receive_tree),
-                    ctx.receive_node(None, ctx.receive_tree)
+                    ctx.receive_node(None, ctx.receive_tree),
+                    ctx.receive_value(None, JavaType)
                 )
 
             if type in ["rewrite.java.tree.Synchronized", "org.openrewrite.java.tree.J$Synchronized"]:
@@ -1329,6 +1359,14 @@ class JavaReceiver(Receiver):
                     ctx.receive_value(None, str)
                 )
 
+            if type in ["rewrite.java.tree.Erroneous", "org.openrewrite.java.tree.J$Erroneous"]:
+                return Erroneous(
+                    ctx.receive_value(None, UUID),
+                    ctx.receive_node(None, JavaReceiver.receive_space),
+                    ctx.receive_node(None, ctx.receive_markers),
+                    ctx.receive_value(None, str)
+                )
+
             raise NotImplementedError("No factory method for type: " + type)
 
     @classmethod
