@@ -1,15 +1,14 @@
 from __future__ import annotations
 
 import weakref
-from dataclasses import dataclass, replace
+from abc import abstractmethod, ABC
+from dataclasses import dataclass, field, replace
 from enum import Enum, auto
-from functools import cached_property
-from typing import List, Optional, Protocol, TypeVar, Generic, ClassVar, Dict, runtime_checkable, Any, cast, \
-    TYPE_CHECKING
+from typing import List, Optional, TypeVar, Generic, ClassVar, Dict, Any, TYPE_CHECKING, Iterable, cast
 from uuid import UUID
 
-from rewrite import Tree, SourceFile, TreeVisitor
 from rewrite import Markers
+from rewrite import Tree, SourceFile, TreeVisitor
 
 if TYPE_CHECKING:
     from .visitor import JavaVisitor
@@ -17,15 +16,20 @@ if TYPE_CHECKING:
 P = TypeVar('P')
 
 
-@runtime_checkable
-class J(Tree, Protocol):
+class J(Tree):
     @property
+    @abstractmethod
     def prefix(self) -> Space:
         ...
 
+    @abstractmethod
     def with_prefix(self, prefix: Space) -> 'J':
         ...
 
+    def is_acceptable(self, v: TreeVisitor[Any, P], p: P) -> bool:
+        from .visitor import JavaVisitor
+        return isinstance(v, JavaVisitor)
+
     def accept(self, v: TreeVisitor[Any, P], p: P) -> Optional[Any]:
         from .visitor import JavaVisitor
         return self.accept_java(v.adapt(J, JavaVisitor), p)
@@ -35,8 +39,9 @@ class J(Tree, Protocol):
 
 
 @dataclass(frozen=True)
-class Comment(Protocol):
+class Comment(ABC):
     @property
+    @abstractmethod
     def multiline(self) -> bool:
         ...
 
@@ -101,8 +106,8 @@ class Space:
     _whitespace: Optional[str]
 
     @property
-    def whitespace(self) -> Optional[str]:
-        return self._whitespace
+    def whitespace(self) -> str:
+        return self._whitespace if self._whitespace is not None else ""
 
     def with_whitespace(self, whitespace: Optional[str]) -> Space:
         return self if whitespace is self._whitespace else replace(self, _whitespace=whitespace)
@@ -110,25 +115,67 @@ class Space:
     def is_empty(self) -> bool:
         return len(self._comments) == 0 and (self._whitespace is None or self._whitespace == '')
 
-    EMPTY: ClassVar[Space] = None
-    SINGLE_SPACE: ClassVar[Space] = None
+    @classmethod
+    def first_prefix(cls, trees: Optional[Iterable[J]]) -> Space:
+        return Space.EMPTY if trees is None or not trees else next(iter(trees)).prefix
+
+    @classmethod
+    def format_first_prefix(cls, trees: List[J2], prefix: Space) -> List[J2]:
+        if trees and next(iter(trees)).prefix != prefix:
+            formatted_trees = list(trees)
+            formatted_trees[0] = cast(J2, formatted_trees[0].with_prefix(prefix))
+            return formatted_trees
+        return trees
+
+    @property
+    def indent(self) -> str:
+        """
+        The indentation after the last newline of either the last comment's suffix
+        or the global whitespace if no comments exist.
+        """
+        return self._get_whitespace_indent(self.last_whitespace)
+
+    @property
+    def last_whitespace(self) -> str:
+        """
+        The raw suffix from the last comment if it exists, otherwise the global
+        whitespace (or empty string if whitespace is None).
+        """
+        if self._comments:
+            return self._comments[-1].suffix
+        return self._whitespace if self._whitespace is not None else ""
+
+    @staticmethod
+    def _get_whitespace_indent(whitespace: Optional[str]) -> str:
+        """
+        A helper method that extracts everything after the last newline character
+        in `whitespace`. If no newline is present, returns `whitespace` as-is.
+        If the last newline is at the end, returns an empty string.
+        """
+        if not whitespace:
+            return ""
+        last_newline = whitespace.rfind('\n')
+        return whitespace if last_newline == -1 else whitespace[last_newline + 1:]
+
+    EMPTY: ClassVar[Space]
+    SINGLE_SPACE: ClassVar[Space]
 
     class Location(Enum):
-        ANY = auto()
         ANNOTATED_TYPE_PREFIX = auto()
+        ANNOTATIONS = auto()
         ANNOTATION_ARGUMENTS = auto()
         ANNOTATION_ARGUMENT_SUFFIX = auto()
-        ANNOTATIONS = auto()
         ANNOTATION_PREFIX = auto()
+        ANY = auto()
         ARRAY_ACCESS_PREFIX = auto()
         ARRAY_INDEX_SUFFIX = auto()
         ARRAY_TYPE_PREFIX = auto()
-        ASSERT_PREFIX = auto()
         ASSERT_DETAIL = auto()
         ASSERT_DETAIL_PREFIX = auto()
+        ASSERT_PREFIX = auto()
         ASSIGNMENT = auto()
-        ASSIGNMENT_OPERATION_PREFIX = auto()
         ASSIGNMENT_OPERATION_OPERATOR = auto()
+        ASSIGNMENT_OPERATION_PREFIX = auto()
         ASSIGNMENT_PREFIX = auto()
         BINARY_OPERATOR = auto()
         BINARY_PREFIX = auto()
@@ -137,9 +184,9 @@ class Space:
         BLOCK_STATEMENT_SUFFIX = auto()
         BREAK_PREFIX = auto()
         CASE = auto()
-        CASE_PREFIX = auto()
         CASE_BODY = auto()
         CASE_EXPRESSION = auto()
+        CASE_PREFIX = auto()
         CASE_SUFFIX = auto()
         CATCH_ALTERNATIVE_SUFFIX = auto()
         CATCH_PREFIX = auto()
@@ -149,8 +196,8 @@ class Space:
         COMPILATION_UNIT_PREFIX = auto()
         CONTINUE_PREFIX = auto()
         CONTROL_PARENTHESES_PREFIX = auto()
-        DIMENSION_PREFIX = auto()
         DIMENSION = auto()
+        DIMENSION_PREFIX = auto()
         DIMENSION_SUFFIX = auto()
         DO_WHILE_PREFIX = auto()
         ELSE_PREFIX = auto()
@@ -158,6 +205,7 @@ class Space:
         ENUM_VALUE_PREFIX = auto()
         ENUM_VALUE_SET_PREFIX = auto()
         ENUM_VALUE_SUFFIX = auto()
+        ERRONEOUS_PREFIX = auto()
         EXPRESSION_PREFIX = auto()
         EXTENDS = auto()
         FIELD_ACCESS_NAME = auto()
@@ -177,9 +225,8 @@ class Space:
         IF_PREFIX = auto()
         IF_THEN_SUFFIX = auto()
         IMPLEMENTS = auto()
-        IMPORT_ALIAS_PREFIX = auto()
-        PERMITS = auto()
         IMPLEMENTS_SUFFIX = auto()
+        IMPORT_ALIAS_PREFIX = auto()
         IMPORT_PREFIX = auto()
         IMPORT_SUFFIX = auto()
         INSTANCEOF_PREFIX = auto()
@@ -196,9 +243,9 @@ class Space:
         MEMBER_REFERENCE_CONTAINING = auto()
         MEMBER_REFERENCE_NAME = auto()
         MEMBER_REFERENCE_PREFIX = auto()
+        METHOD_DECLARATION_DEFAULT_VALUE = auto()
         METHOD_DECLARATION_PARAMETERS = auto()
         METHOD_DECLARATION_PARAMETER_SUFFIX = auto()
-        METHOD_DECLARATION_DEFAULT_VALUE = auto()
         METHOD_DECLARATION_PREFIX = auto()
         METHOD_INVOCATION_ARGUMENTS = auto()
         METHOD_INVOCATION_ARGUMENT_SUFFIX = auto()
@@ -223,6 +270,7 @@ class Space:
         PARAMETERIZED_TYPE_PREFIX = auto()
         PARENTHESES_PREFIX = auto()
         PARENTHESES_SUFFIX = auto()
+        PERMITS = auto()
         PERMITS_SUFFIX = auto()
         PRIMITIVE_PREFIX = auto()
         RECORD_STATE_VECTOR = auto()
@@ -231,8 +279,8 @@ class Space:
         STATEMENT_PREFIX = auto()
         STATIC_IMPORT = auto()
         STATIC_INIT_SUFFIX = auto()
-        SWITCH_PREFIX = auto()
         SWITCH_EXPRESSION_PREFIX = auto()
+        SWITCH_PREFIX = auto()
         SYNCHRONIZED_PREFIX = auto()
         TERNARY_FALSE = auto()
         TERNARY_PREFIX = auto()
@@ -271,68 +319,229 @@ Space.EMPTY = Space([], '')
 Space.SINGLE_SPACE = Space([], ' ')
 
 
-@runtime_checkable
-class JavaSourceFile(SourceFile, Protocol):
+@dataclass
+class CoordinateBuilder:
+    tree: J
+
+    def before(self, loc: Space.Location) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc, JavaCoordinates.Mode.BEFORE)
+
+    def after(self, loc: Space.Location) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc, JavaCoordinates.Mode.AFTER)
+
+    def replace(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc, JavaCoordinates.Mode.REPLACE)
+
+
+@dataclass
+class _ExpressionCoordinateBuilder(CoordinateBuilder):
+    def after(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc or Space.Location.EXPRESSION_PREFIX, JavaCoordinates.Mode.AFTER)
+
+    def before(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc or Space.Location.EXPRESSION_PREFIX, JavaCoordinates.Mode.BEFORE)
+
+    def replace(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc or Space.Location.EXPRESSION_PREFIX, JavaCoordinates.Mode.REPLACE)
+
+
+@dataclass
+class _StatementCoordinateBuilder(CoordinateBuilder):
+    def after(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc or Space.Location.STATEMENT_PREFIX, JavaCoordinates.Mode.AFTER)
+
+    def before(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc or Space.Location.STATEMENT_PREFIX, JavaCoordinates.Mode.BEFORE)
+
+    def replace(self, loc: Optional[Space.Location] = None) -> JavaCoordinates:
+        return JavaCoordinates(self.tree, loc or Space.Location.STATEMENT_PREFIX, JavaCoordinates.Mode.REPLACE)
+
+
+@dataclass
+class _BlockCoordinateBuilder(_StatementCoordinateBuilder):
+    def first_statement(self) -> JavaCoordinates:
+        if not self.tree.statements:
+            return self.last_statement()
+        return self.tree.statements[0].get_coordinates().before()
+
+    def last_statement(self) -> JavaCoordinates:
+        return self.before(Space.Location.BLOCK_END)
+
+
+CoordinateBuilder.Expression = _ExpressionCoordinateBuilder  # type: ignore
+CoordinateBuilder.Statement = _StatementCoordinateBuilder  # type: ignore
+CoordinateBuilder.Block = _BlockCoordinateBuilder  # type: ignore
+
+
+@dataclass
+class JavaCoordinates:
+    tree: J
+    loc: Space.Location
+    mode: Mode
+
+    def is_replacement(self) -> bool:
+        return self.mode == JavaCoordinates.Mode.REPLACE
+
+    class Mode(Enum):
+        AFTER = 0,
+        BEFORE = 1,
+        REPLACE = 2,
+
+
+class JavaSourceFile(J, SourceFile):
     pass
 
 
-@runtime_checkable
-class Expression(J, Protocol):
+class Expression(J):
+    def get_coordinates(self) -> CoordinateBuilder.Expression:  # type: ignore
+        return CoordinateBuilder.Expression(self)  # type: ignore
+
+
+class Statement(J):
+    def get_coordinates(self) -> CoordinateBuilder.Statement:  # type: ignore
+        return CoordinateBuilder.Statement(self)  # type: ignore
+
+
+class TypedTree(J):
     pass
 
 
-@runtime_checkable
-class Statement(J, Protocol):
+class NameTree(TypedTree):
     pass
 
 
-@runtime_checkable
-class TypedTree(J, Protocol):
+class TypeTree(NameTree):
     pass
 
 
-@runtime_checkable
-class NameTree(TypedTree, Protocol):
+class Loop(Statement):
     pass
 
 
-@runtime_checkable
-class TypeTree(NameTree, Protocol):
+class MethodCall(Expression):
     pass
 
 
-@runtime_checkable
-class Loop(Statement, Protocol):
-    pass
-
-
-@runtime_checkable
-class MethodCall(Expression, Protocol):
-    pass
-
-
-class JavaType(Protocol):
+class JavaType(ABC):
     class FullyQualified:
+        class Kind(Enum):
+            Class = 0
+            Enum = 1
+            Interface = 2
+            Annotation = 3
+            Record = 4
+
+    class Unknown(FullyQualified):
         pass
 
-    class Class:
+    class Class(FullyQualified):
+        _flags_bit_map: int
+        _fully_qualified_name: str
+        _kind: FullyQualified.Kind
+        _type_parameters: Optional[List[JavaType]]
+        _supertype: Optional[JavaType.FullyQualified]
+        _owning_class: Optional[JavaType.FullyQualified]
+        _annotations: Optional[List[JavaType.FullyQualified]]
+        _interfaces: Optional[List[JavaType.FullyQualified]]
+        _members: Optional[List[JavaType.Variable]]
+        _methods: Optional[List[JavaType.Method]]
+
+    class ShallowClass(Class):
         pass
 
-    class Parameterized:
-        pass
+    class Parameterized(FullyQualified):
+        _type: JavaType.FullyQualified
+        _type_parameters: Optional[List[JavaType]]
 
-    class Primitive:
-        pass
+    class GenericTypeVariable:
+        class Variance(Enum):
+            Invariant = 0
+            Covariant = 1
+            Contravariant = 2
 
+    class Primitive(Enum):
+        Boolean = 0
+        Byte = 1
+        Char = 2
+        Double = 3
+        Float = 4
+        Int = 5
+        Long = 6
+        Short = 7
+        Void = 8
+        String = 9
+        None_ = 10
+        Null = 11
+
+        @classmethod
+        def _missing_(cls, value):
+            if value is None:
+                return cls.None_
+            return super()._missing_(value)
+
+    @dataclass
     class Method:
-        pass
+        _flags_bit_map: int = field(default=0)
+        _declaring_type: Optional[JavaType.FullyQualified] = field(default=None)
+        _name: str = field(default="")
+        _return_type: Optional[JavaType] = field(default=None)
+        _parameter_names: Optional[List[str]] = field(default=None)
+        _parameter_types: Optional[List[JavaType]] = field(default=None)
+        _thrown_exceptions: Optional[List[JavaType]] = field(default=None)
+        _annotations: Optional[List[JavaType.FullyQualified]] = field(default=None)
+        _default_value: Optional[List[str]] = field(default=None)
+        _declared_formal_type_names: Optional[List[str]] = field(default=None)
+
+        @property
+        def flags_bit_map(self) -> int:
+            return self._flags_bit_map
+
+        @property
+        def declaring_type(self) -> Optional[JavaType.FullyQualified]:
+            return self._declaring_type
+
+        @property
+        def name(self) -> str:
+            return self._name
+
+        @property
+        def return_type(self) -> Optional[JavaType]:
+            return self._return_type
+
+        @property
+        def parameter_names(self) -> Optional[List[str]]:
+            return self._parameter_names
+
+        @property
+        def parameter_types(self) -> Optional[List[JavaType]]:
+            return self._parameter_types
+
+        @property
+        def thrown_exceptions(self) -> Optional[List[JavaType]]:
+            return self._thrown_exceptions
+
+        @property
+        def annotations(self) -> Optional[List[JavaType.FullyQualified]]:
+            return self._annotations
+
+        @property
+        def default_value(self) -> Optional[List[str]]:
+            return self._default_value
+
+        @property
+        def declared_formal_type_names(self) -> Optional[List[str]]:
+            return self._declared_formal_type_names
 
     class Variable:
         pass
 
+    class Array:
+        pass
+
 
 T = TypeVar('T')
 J2 = TypeVar('J2', bound=J)
+J3 = TypeVar('J3', bound=J)
 
 
 @dataclass(frozen=True)
@@ -504,23 +713,23 @@ class JLeftPadded(Generic[T]):
 
 
 @dataclass(frozen=True)
-class JContainer(Generic[T]):
+class JContainer(Generic[J2]):
     _before: Space
 
     @property
     def before(self) -> Space:
         return self._before
 
-    def with_before(self, before: Space) -> JContainer[T]:
+    def with_before(self, before: Space) -> JContainer[J2]:
         return self if before is self._before else replace(self, _before=before)
 
-    _elements: List[JRightPadded[T]]
+    _elements: List[JRightPadded[J2]]
 
     @property
-    def elements(self) -> List[T]:
+    def elements(self) -> List[J2]:
         return JRightPadded.get_elements(self._elements)
 
-    def with_elements(self, elements: List[T]) -> JContainer[T]:
+    def with_elements(self, elements: List[J2]) -> JContainer[J2]:
         return self.padding.with_elements(JRightPadded.with_elements(self._elements, elements))
 
     _markers: Markers
@@ -529,25 +738,25 @@ class JContainer(Generic[T]):
     def markers(self) -> Markers:
         return self._markers
 
-    def with_markers(self, markers: Markers) -> JContainer[T]:
+    def with_markers(self, markers: Markers) -> JContainer[J2]:
         return self if markers is self._markers else replace(self, _markers=markers)
 
     @dataclass
-    class PaddingHelper:
-        _t: JContainer[T]
+    class PaddingHelper(Generic[J3]):
+        _t: JContainer[J3]
 
         @property
-        def elements(self) -> List[JRightPadded[T]]:
+        def elements(self) -> List[JRightPadded[J3]]:
             return self._t._elements
 
-        def with_elements(self, elements: List[JRightPadded[T]]) -> JContainer[T]:
+        def with_elements(self, elements: List[JRightPadded[J3]]) -> JContainer[J3]:
             return self._t if self._t._elements is elements else JContainer(self._t._before, elements, self._t._markers)
 
-    _padding: weakref.ReferenceType[JContainer.PaddingHelper] = None
+    _padding: Optional[weakref.ReferenceType[JContainer.PaddingHelper[J2]]] = None
 
     @property
-    def padding(self) -> JContainer.PaddingHelper:
-        p: JContainer.PaddingHelper
+    def padding(self) -> JContainer.PaddingHelper[J2]:
+        p: Optional[JContainer.PaddingHelper[J2]]
         if self._padding is None:
             p = JContainer.PaddingHelper(self)
             object.__setattr__(self, '_padding', weakref.ref(p))
@@ -565,16 +774,16 @@ class JContainer(Generic[T]):
         if elements is None or elements == []:
             return None
         if before is None:
-            return JContainer(Space.EMPTY, elements, Markers.EMPTY)
+            return JContainer(Space.EMPTY, JRightPadded.with_elements([], elements), Markers.EMPTY)
         return before.padding.with_elements(JRightPadded.with_elements(before._elements, elements))
 
-    _EMPTY = None
+    _EMPTY: Optional[JContainer[J]] = None
 
     @classmethod
-    def empty(cls) -> JContainer[T]:
+    def empty(cls) -> JContainer[J2]:
         if cls._EMPTY is None:
             cls._EMPTY = JContainer(Space.EMPTY, [], Markers.EMPTY)
-        return cls._EMPTY
+        return cls._EMPTY  # type: ignore
 
     class Location(Enum):
         ANNOTATION_ARGUMENTS = (Space.Location.ANNOTATION_ARGUMENTS, JRightPadded.Location.ANNOTATION_ARGUMENT)
