@@ -1,6 +1,6 @@
 from typing import Optional, TypeVar, TYPE_CHECKING
 
-from rewrite.java.tree import J, JContainer, JRightPadded, JLeftPadded, Space
+from rewrite.java.tree import extensions, J, JContainer, JRightPadded, JLeftPadded, Space
 from .support_types import PyContainer, PyRightPadded, PyLeftPadded, PySpace
 
 if TYPE_CHECKING:
@@ -13,22 +13,22 @@ J2 = TypeVar('J2', bound=J)
 # noinspection PyUnusedLocal
 def visit_container(v: 'PythonVisitor', container: Optional[JContainer[J2]],
                     loc: PyContainer.Location, p) -> Optional[JContainer[J2]]:
-    return None
+    return extensions.visit_container(v, container, JContainer.Location.LANGUAGE_EXTENSION, p)
 
 
 # noinspection PyUnusedLocal
 def visit_right_padded(v: 'PythonVisitor', right: Optional[JRightPadded[T]],
                        loc: PyRightPadded.Location, p) -> Optional[JRightPadded[T]]:
-    return None
+    return extensions.visit_right_padded(v, right, JRightPadded.Location.LANGUAGE_EXTENSION, p)
 
 
 # noinspection PyUnusedLocal
 def visit_left_padded(v: 'PythonVisitor', left: Optional[JLeftPadded[T]],
                       loc: PyLeftPadded.Location, p) -> Optional[JLeftPadded[T]]:
-    return None
+    return extensions.visit_left_padded(v, left, JLeftPadded.Location.LANGUAGE_EXTENSION, p)
 
 
 # noinspection PyUnusedLocal
 def visit_space(v: 'PythonVisitor', space: Optional[Space],
                 loc: PySpace.Location, p):
-    return None
+    return extensions.visit_space(v, space, Space.Location.LANGUAGE_EXTENSION, p)
