@@ -2,7 +2,7 @@ from __future__ import annotations
 
 from dataclasses import replace, dataclass
 from enum import Enum, auto
-from typing import Protocol, TypeVar, runtime_checkable, Any, Optional, TYPE_CHECKING
+from typing import TypeVar, Any, Optional, TYPE_CHECKING
 
 from rewrite import TreeVisitor, Markers
 from rewrite.java.tree import J
@@ -14,8 +14,7 @@ if TYPE_CHECKING:
 P = TypeVar('P')
 
 
-@runtime_checkable
-class Py(J, Protocol):
+class Py(J):
     def accept(self, v: TreeVisitor[Any, P], p: P) -> Optional[Any]:
         from .visitor import PythonVisitor
         return self.accept_python(v.adapt(Py, PythonVisitor), p)
@@ -23,6 +22,10 @@ class Py(J, Protocol):
     def accept_python(self, v: 'PythonVisitor[P]', p: P) -> Optional['J']:
         ...
 
+    def is_acceptable(self, v: TreeVisitor[Any, P], p: P) -> bool:
+        from .visitor import PythonVisitor
+        return isinstance(v, PythonVisitor)
+
 
 class PySpace:
     class Location(Enum):
@@ -60,6 +63,7 @@ class PySpace:
         FORMATTED_STRING_PREFIX = auto()
         FORMATTED_STRING_VALUE_DEBUG_SUFFIX = auto()
         FORMATTED_STRING_VALUE_PREFIX = auto()
+        FOR_LOOP_BODY_SUFFIX = auto()
         FOR_LOOP_ITERABLE = auto()
         FOR_LOOP_PREFIX = auto()
         KEY_VALUE_PREFIX = auto()
@@ -114,6 +118,7 @@ class PyRightPadded:
         DICT_LITERAL_ELEMENT = PySpace.Location.DICT_LITERAL_ELEMENT_SUFFIX
         FORMATTED_STRING_VALUE_DEBUG = PySpace.Location.FORMATTED_STRING_VALUE_DEBUG_SUFFIX
         FORMATTED_STRING_VALUE_EXPRESSION = PySpace.Location.FORMATTED_STRING_VALUE_PREFIX
+        FOR_LOOP_BODY = PySpace.Location.FOR_LOOP_BODY_SUFFIX
         KEY_VALUE_KEY = PySpace.Location.DICT_ENTRY_KEY_SUFFIX
         KEY_VALUE_KEY_SUFFIX = PySpace.Location.KEY_VALUE_SUFFIX
         MATCH_CASE_PATTERN_CHILD = PySpace.Location.MATCH_PATTERN_ELEMENT_SUFFIX
